\treat.xpl	28-Sep-2020	Loren Blaney	loren.blaney@gmail.com
\Trick-or-Treat Game
\Compile with 16-bit XPL (x, xn, or xx).

include c:\cxpl\codesi;		\intrinsic 'code' declarations

def	IntSize = 2,		\number of bytes in an integer
	Cols=21, Rows=13,	\dimensions of neighborhoods (map objects)
	StkMax = 30000;		\stack for saving & undoing 30000/8 = 3750 moves
char	Stack(StkMax),
	Map(Cols, Rows);	\map of neighborhood
def	\MapObject\ Road, Lawn, Man, Orb, Treat;

char	Image,			\contains seven 15x14 sprites on 16x14 grid, etc
	Palette,		\color palette array for sprite images
	TreatData;		\data file describing levels (a big string)

int	Best,			\array of fewest step counts for each level
	InxData,		\index into TreatData string
	Key,			\keystroke's ASCII value or scan code
	Level,			\level (i.e. neighborhood) number (1..MaxLevel)
	MaxLevel,		\maximum level ('hood) number
	MoveCtr,		\move counter (= number of arrow keystrokes)
	ManX, ManY,		\man's (Herbie/ghost) location in Map
	NewX, NewY,		\possible new man location
	OrbX, OrbY,		\possible new orb (pumpkin) location
	StkPtr,			\stack pointer (for Push and Pop routines)
	Steps;			\array of step counts (MoveCtr) for each level

def	UpArrow=$48, DnArrow=$50, LfArrow=$4B, RtArrow=$4D;    \key scan codes
def	Black, Blue, Green, Cyan, Red, Magenta, Brown, White,  \attribute colors
	Gray, LBlue, LGreen, LCyan, LRed, LMagenta, Yellow, BWhite; \EGA palette

\Sprite images representing the objects
def	SprW=15, SprH=14;	\sprite dimensions (pixels)
char	SprLawn(2+SprW*SprH),
	SprRoad(2+SprW*SprH),
	SprMan(2+SprW*SprH),
	SprOrb(2+SprW*SprH),
	SprTreat(2+SprW*SprH),
	SprWeed(2+SprW*SprH),
	SprSpider(2+SprW*SprH),
	SprHood(2+4*8*8),	\"hood"
	SprSteps(2+5*8*8),	\"steps"
	SprSlash(2+1*8*8),	\"/"
	SprDigit(10, 2+8*8);	\0 1 2 3 4 5 6 7 8 9

def	HoodX=(320-Cols*SprW)/2, HoodY=10;	\offset for centering map

\------------------------ Routines for DOS version -----------------------------

int	CpuReg;			\to access CPU registers

func	CallInt(Int, AX, BX, CX, DX, BP, DS, ES); \Call software interrupt
int	Int, AX, BX, CX, DX, BP, DS, ES; \(unused arguments need not be passed)
[CpuReg(0):= AX;
CpuReg(1):= BX;
CpuReg(2):= CX;
CpuReg(3):= DX;
CpuReg(6):= BP;
CpuReg(9):= DS;
CpuReg(11):= ES;
SoftInt(Int);
return CpuReg(0);		\return contents of AX register
];	\CallInt



proc	KBRepeat(Rate);		\Set keyboard auto-repeat rate
int	Rate;			\use 12 for 10 chars per second
CallInt($16, $0305, $0100+Rate); \standard delay is 0.5 sec. before auto-repeat



proc	SetPalette(Reg, R, G, B); \Set a palette color register
int	Reg, R, G, B;
[port($3C8):= Reg;		\set starting color register
 port($3C9):= R>>2;		\Red (8 bits to 6 bits)
 port($3C9):= G>>2;
 port($3C9):= B>>2;
];	\SetPalette



proc	WaitForVSync;		\Wait for start of next vertical blank, actually
int	SR;			\status register
[SR:= Peek($40, $63) + $306;	\port address of status register
if (SR & $39A) # $39A then return; \prevents waiting forever
while PIn(SR, 0) & $08 do [];	\wait for no vertical blank
repeat until PIn(SR, 0) & $08;	\wait for vertical blank
];	\WaitForVSync

\-------------------------------------------------------------------------------

proc	ReadBest;		\Read best score data from disk file
int	Han, I;
[Trap(false);
Han:= FOpen("treat.dat", 0);
FSet(Han, ^i);
OpenI(3);
if GetErr then			\file doesn't exist, so start a new one
	for I:= 0, MaxLevel-1 do Best(I):= 999
else	[for I:= 0, MaxLevel-1 do Best(I):= IntIn(3);
	if GetErr then		\something's wrong, so start a new file
		for I:= 0, MaxLevel-1 do Best(I):= 999
	];
FClose(Han);			\close file handle
I:= GetErr;			\discard any pending errors
Trap(true);
];	\ReadBest



proc	WriteBest;		\Write best score data to disk file
int	Han, I;
[Han:= FOpen("treat.dat", 1);
FSet(Han, ^o);
OpenO(3);
for I:= 0, MaxLevel-1 do
	[IntOut(3, Best(I));  Crlf(3)];
Close(3);
FClose(Han);
];	\WriteBest



proc	JustOut(Dev, N);	\Output right-justified integer in range 0..9999
int	Dev, N;
[if N <1000 then ChOut(Dev, ^ );
 if N < 100 then ChOut(Dev, ^ );
 if N <  10 then ChOut(Dev, ^ );
IntOut(Dev, N);
];	\JustOut



proc	Exit(Msg);		\Display (error) message and terminate program
char	Msg;
int	I;
[WriteBest;
SetVid(3);			\make sure normal (80x25) text mode is restored
Text(0, Msg);
CrLf(0);

CrLf(0);			\show scores
Text(0, " Scores
 'hood steps / best
");
for I:= 0, MaxLevel-1 do
	[JustOut(0, I+1);  ChOut(0, 9\tab\);
	 JustOut(0, Steps(I));  Text(0, " / ");
	 IntOut(0, Best(I));  Crlf(0);
	];
Crlf(0);
exit;
];	\Exit



proc	Push(B);		\Push byte B onto the Stack
int	B;
[if StkPtr < StkMax then
	[Stack(StkPtr):= B;
	StkPtr:= StkPtr+1;
	];
];	\Push



func	Pop;			\Pop byte from Stack and return it
[if StkPtr > 0 then
	StkPtr:= StkPtr-1;
return Stack(StkPtr);
];	\Pop



proc	FadeIn;	\Slowly increase intensities of palette colors to final values
int	I, J;
char	Pal(256*3);
[for I:= 1 to 64 do		\64 steps take about one second
	[for J:= 0 to 256*3-1 do Pal(J):= Palette(J)*I/64;
	WaitForVSync;
	for J:= 0 to 256-1 do SetPalette(J, Pal(J*3+0), Pal(J*3+1), Pal(J*3+2));
	];
];	\FadeIn



proc	FadeOut; \Slowly decrease intensities of palette colors to black
int	I, J;
char	Pal(256*3);
[for I:= 63 downto 0 do		\64 steps take about one second
	[for J:= 0 to 256*3-1 do Pal(J):= Palette(J)*I/64;
	WaitForVSync;
	for J:= 0 to 256-1 do SetPalette(J, Pal(J*3+0), Pal(J*3+1), Pal(J*3+2));
	];
];	\FadeOut

\-------------------------------------------------------------------------------

proc	LoadMap;		\Load next 'hood level into Map array
int	X, Y, Ch;
[for Y:= 0 to Rows-1 do
    for X:= 0 to Cols-1 do
	[repeat	Ch:= TreatData(InxData);  InxData:= InxData+1;
	until Ch>=$20 & Ch<=$7E & Ch#^>; \skip CR, LF, etc.
	case Ch of		\convert characters to recognizable names
	  ^ :	Map(X, Y):= Road;
	  ^#:	Map(X, Y):= Lawn;
	  ^o:	Map(X, Y):= Orb;
	  ^&:  [Map(X, Y):= Man;  ManX:= X;  ManY:= Y];	\record man's position
	  ^!:	Map(X, Y):= Treat
	other	HexOut(0, Ch);		\show any illegal characters
	];
];	\LoadMap



proc	DrawSprite(X0, Y0, Spr);	\Display a sprite image
int	X0, Y0;	\coordinates of upper-left corner (pixels)
char	Spr;	\address of sprite data
int	X, Y, W, H;
[W:= Spr(0);			\get width and height (in pixels)
 H:= Spr(1);
Spr:= Spr+2;
for Y:= Y0 to Y0+H-1 do
    for X:= X0 to X0+W-1 do
	[if Spr(0) then		\black (0) is transparent; don't plot it
	    Point(X, Y, Spr(0));
	Spr:= Spr+1;
	];
];	\DrawSprite



proc	DrawObj(X, Y, Obj);	\Convert object to its sprite and draw it
int	X, Y, Obj;		\Map cell coordinates and object
[case Obj of
  Road:	Obj:= SprRoad;
  Lawn:	Obj:= if Ran(30) then SprLawn
	      else (if Ran(5) then SprWeed else SprSpider);
  Orb:	Obj:= SprOrb;
  Man:	Obj:= SprMan;
  Treat:Obj:= SprTreat
other	Exit("Illegal object code");
DrawSprite(X*SprW+HoodX, Y*SprH+HoodY, Obj);
];	\DrawObj



int	CurX;		\Cursor position for displaying sprite digits

proc	NumOut(N);	\Display positive integer at CurX,1 using sprite digits
int	N;
int	Q, R;
[Q:= N/10;
R:= rem(0);
if Q then NumOut(Q);	\recurse
DrawSprite(CurX*8, 8*1, @SprDigit(R, 0));
CurX:= CurX+1;
];	\NumOut



proc	ShowSteps;		\Display MoveCtr, etc.
[for CurX:= 0 to 6 do		\blank out current numbers
	DrawSprite(26*8+CurX*SprW, 8*1, SprLawn);
CurX:= 20+6;
NumOut(MoveCtr);		\show new number
CurX:= CurX+1;			\space
DrawSprite(CurX*8, 1*8, SprSlash);
CurX:= CurX+1+1;		\space
NumOut(Best(Level-1));
];	\ShowSteps



proc	DrawMap;		\Show current 'hood (man, orbs, lawn, etc.)
int	X, Y;			\map cell coordinates
[for Y:= 0 to 199/SprH do	\make lawn extend to edge of screen
	[DrawSprite(0,         Y*SprH, SprLawn);
	 DrawSprite(Cols*SprW, Y*SprH, SprLawn);
	];
for X:= 0 to 319/SprW do
	[DrawSprite(X*SprW, 0, SprLawn);
	 DrawSprite(X*SprW, Rows*SprH+HoodY, SprLawn);
	];

for Y:= 0 to Rows-1 do
    for X:= 0 to Cols-1 do
	DrawObj(X, Y, Map(X, Y));

\0123456789012345678901234567890123456789
\      hood 3        steps 23 / 999      

DrawSprite(6*8, 1*8, SprHood);
CurX:= 6+4+1;
NumOut(Level);

DrawSprite(20*8, 1*8, SprSteps);
ShowSteps;
];	\DrawMap

\-------------------------------------------------------------------------------

func	Solved;			\Return 'true' if all treats are gone
int	X, Y;
[for Y:= 0 to Rows-1 do
    for X:= 0 to Cols-1 do
	if Map(X, Y) = Treat then return false;
return true;
];	\Solved



proc	Undo;			\Undo a move by restoring previous state
int	T;
[if StkPtr > 0 then		\there is a move saved on the stack
       [OrbY:= Pop;		\restore what was new location of orb
	OrbX:= Pop;
	NewY:= Pop;		\restore what was new location of man
	NewX:= Pop;
	ManY:= Pop;		\restore location where man is
	ManX:= Pop;
	T:= Pop;
	Map(OrbX, OrbY):= T;	\restore contents where orb might go
	DrawObj(OrbX, OrbY, T);
	T:= Pop;
	Map(NewX, NewY):= T;	\restore contents at new man location
	DrawObj(NewX, NewY, T);
	Map(ManX, ManY):= Man;	\restore man
	DrawObj(ManX, ManY, Man);
	MoveCtr:= MoveCtr-1;	\undo move count
	ShowSteps;
       ];
];	\Undo



proc	DoMove;			\Arrow keys move man and possibly an orb
int	N;

	proc	MoveMan;	\Move man to New and remove man from ManX,ManY
	[Map(ManX, ManY):= Road;
	DrawObj(ManX, ManY, Road);
	Map(NewX, NewY):= Man;
	DrawObj(NewX, NewY, Man);
	ManX:= NewX;  ManY:= NewY;	\set man to new location
	];	\MoveMan

[NewX:= ManX;  NewY:= ManY;
OrbX:= ManX;  OrbY:= ManY;
case Key of	\determine possible new locations for man and orb(s)
  UpArrow:	[NewY:= ManY-1;
		repeat OrbY:= OrbY-1 until Map(OrbX, OrbY) # Orb;
		];
  LfArrow:	[NewX:= ManX-1;
		repeat OrbX:= OrbX-1 until Map(OrbX, OrbY) # Orb;
		];
  RtArrow:	[NewX:= ManX+1;
		repeat OrbX:= OrbX+1 until Map(OrbX, OrbY) # Orb;
		];
  DnArrow:	[NewY:= ManY+1;
		repeat OrbY:= OrbY+1 until Map(OrbX, OrbY) # Orb;
		]
other		[];

\Save state so move can be undone
Push(Map(NewX, NewY));	\save where man might go (path, treat, orb, lawn)
Push(Map(OrbX, OrbY));	\save where orb might go (path, or treat, lawn)
Push(ManX);		\save location of man
Push(ManY);
Push(NewX);		\save possible new location of man
Push(NewY);
Push(OrbX);		\save possible new location of orb
Push(OrbY);
MoveCtr:= MoveCtr+1;	\bump up move counter (even if move is illegal)
ShowSteps;

case Map(NewX, NewY) of
  Road:	MoveMan;
  Orb:	if Map(OrbX, OrbY) = Road then	\attempting to move orb
		[Map(OrbX, OrbY):= Orb;	\move orb to new location
		DrawObj(OrbX, OrbY, Orb);
		MoveMan;
		];
  Treat:[MoveMan;			\move man to new location
	for N:= 1 to 3 do		\flash treat and make rising sound
		[Sound(false, 1, 1);
		DrawObj(NewX, NewY, Treat);
		Sound(true, 1, 500*(4-N));
		DrawObj(NewX, NewY, Man);
		];
	]
other	[];				\attempted to move onto lawn
];	\DoMove

\-------------------------------------------------------------------------------

proc	GetStartingLevel;	\Show intro screen and get starting level number
int	N, Ch;
[ChOut(0, $0C\FF\);		\erase screen with form feed
Cursor(31, 2);  Attrib(Yellow);
RawText(6, "¿      ¿ ");
Cursor(32, 3);
RawText(6, "rick or reat ");
Cursor(20, 5);  Attrib(Brown);
Text(6, "Little Herbie has donned his ghost
                    costume and is eager for a night of
                    trick-or-treating. Can you guide him
                    through the neighborhoods using the
                    arrow keys and collect all the treats?

                    An old goblin (who wants the treats for
                    himself) has blocked some paths with
                    pumpkins, but they can be pushed out of
                    the way. Be careful though, if you push
                    one into a corner, you might need to
                    undo the move with the Backspace key.

                    Oh yeah, please keep off the grass!");

loop	[Cursor(22, 20);  Attrib(Yellow);
	Text(6, "Pick starting neighborhood (1-");
	IntOut(6, MaxLevel);  Text(6, "): ");
	Ch:= ChIn(6);
	if Ch>=^1 & Ch<=MaxLevel+^0 then quit;
	if Ch=$1B\Esc\ ! Ch=$03\Ctrl+C\ then Exit(" ");
	Cursor(22, 22);
	Text(6, "Please select 1 thru ");  IntOut(6, MaxLevel);
	];
Level:= Ch - ^0;

\Skip to selected level in input file
InxData:= 0;
for N:= 1 to Level do
	repeat Ch:= TreatData(InxData); InxData:= InxData+1 until Ch = ^>;
];	\GetStartingLevel



proc	GetSprite(X0, Y0, W, H, Spr);	\Load sprite array from Image array
int	X0, Y0,	\coordinates in Image of upper-left corner of sprite
	W, H;	\width and height of sprite (pixels)
char	Spr;	\address of array in which to load sprite data: Spr(2+W*H)
int	X, Y;
[Spr(0):= W;
 Spr(1):= H;
Spr:= Spr+2;
for Y:= Y0 to Y0+H-1 do
    for X:= X0 to X0+W-1 do
	[Spr(0):= Image(X+Y*96);	\Image is 96x32 bytes
	Spr:= Spr+1;
	];
];	\GetSprite



proc	LoadSprites;		\Load sprite arrays from Image data
int	D;
[include treatspr;		\constant arrays containing Image data
GetSprite(0*16, 0, SprW, SprH, SprLawn);
GetSprite(1*16, 0, SprW, SprH, SprRoad);
GetSprite(2*16, 0, SprW, SprH, SprMan);
GetSprite(3*16, 0, SprW, SprH, SprOrb);
GetSprite(4*16, 0, SprW, SprH, SprTreat);
GetSprite(5*16, 0, SprW, SprH, SprWeed);
GetSprite(5*16,16, SprW, SprH, SprSpider);
for D:= 0 to 9 do
	GetSprite(D*8, 16, 8, 8, @SprDigit(D, 0));
GetSprite(0*8, 24, 1*8, 8, SprSlash);
GetSprite(1*8, 24, 4*8, 8, SprHood);
GetSprite(5*8, 24, 5*8, 8, SprSteps);
];	\LoadSprites

\===============================================================================

int	I;
[include treatdta;		\big string containing level data ('hood maps)
Best:=  Reserve(IntSize*MaxLevel);
Steps:= Reserve(IntSize*MaxLevel);
for I:= 0, MaxLevel-1 do Steps(I):= 0;
CpuReg:= GetReg;			\to access CPU registers for DOS
TrapC(true);				\don't let Ctrl+C bypass Exit routine
KBRepeat(12);				\set to old standard of 10 chars/sec
ReadBest;
LoadSprites;
GetStartingLevel;
SetVid($13);				\set 320x200x8 graphics (takes a second)
FadeOut;				\prepare for FadeIn (and delay 1 second)

repeat	[StkPtr:= 0;			\play one level per repeat loop
	MoveCtr:= 0;			\initialize StkPtr for Undo command
	LoadMap;
	DrawMap;
	FadeIn;
	loop	[Key:= ChIn(1);		\one move per loop
		case Key of
		  $1B:	[FadeOut;  Exit(" ")];	\Esc
		  $08:	Undo;		\Backspace = undo previous move
		  UpArrow, DnArrow, LfArrow, RtArrow: DoMove
		other	[];		\ignore 0 prefix and illegal keystrokes

		if Solved then
			[Steps(Level-1):= MoveCtr;	\record number of steps
			if MoveCtr < Best(Level-1) then Best(Level-1):= MoveCtr;
			ShowSteps;	\show best score
			Key:= ChIn(1);	\wait for keystroke
			OpenI(0);	\discard possible arrow key scan code
			quit;		\[OpenI(1) doesn't do the job]
			];
		];	\move loop
	FadeOut;
	Level:= Level+1;		\next neighborhood
	];
until Level > MaxLevel;
Exit("Last 'hood cleaned out!");
]
